@@ -2,15 +2,14 @@
 import os
 import re
 import sys
+import pickle
 import shutil
 import logging
-import tarfile
 import platform
 from zipfile import ZipFile
 from stat import S_IRUSR, S_IXUSR
 from contextlib import contextmanager
-from subprocess import run, PIPE, Popen
-from json.decoder import JSONDecodeError
+from subprocess import run, PIPE, STDOUT, Popen
 
 ### Logging ###
 import logzero
@@ -23,9 +22,7 @@ import click
 import requests
 
 ### Parsing ###
-import jsonpickle
 from lxml import html
-from bs4 import BeautifulSoup
 from semantic_version import Version
 
 ### Display ###
@@ -35,10 +32,6 @@ from tqdm import tqdm
 semver = re.compile(r"^((0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(-(0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(\.(0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*)?(\+[0-9a-zA-Z-]+(\.[0-9a-zA-Z-]+)*)?)/?$")  # noqa: E501
 
 
-CACHE = os.path.join(os.path.expanduser("~"), ".renutil")
-INSTANCE_REGISTRY = os.path.join(CACHE, "index.json")
-
-
 class AliasedGroup(click.Group):
 
     def get_command(self, ctx, cmd_name):
@@ -104,6 +97,82 @@ class RenpyRelease(ComparableVersion):
         return "RenpyRelease(version={}, url='{}')".format(self.version, self.url)
 
 
+class Registry():
+
+    def __init__(self, filename):
+        self.filename = filename
+        self.instances = []
+
+    def load(self):
+        if not os.path.isfile(self.filename):
+            self.persist()
+        with open(self.filename, "rb") as f:
+            self.instances = pickle.load(f)
+
+    def persist(self):
+        with open(self.filename, "wb") as f:
+            pickle.dump(self.instances, f)
+
+    def clear(self):
+        self.instances = []
+        self.persist()
+
+    def add_instance(self, instance):
+        self.load()
+        if instance in self.instances:
+            return
+
+        self.instances.append(instance)
+
+        self.persist()
+
+    def remove_instance(self, instance):
+        self.load()
+        if instance not in self.instances:
+            return
+
+        for i, inst in enumerate(self.instances):
+            if inst.version == instance.version:
+                del self.instances[i]
+
+        self.persist()
+
+    def get_instance(self, version):
+        self.load()
+        if isinstance(version, str):
+            try:
+                version = Version(version)
+            except ValueError:
+                return None
+
+        for instance in self.instances:
+            if instance.version == version:
+                return instance
+
+        return None
+
+    def installed(self, version):
+        self.load()
+        if isinstance(version, str):
+            try:
+                version = Version(version)
+            except ValueError:
+                return False
+
+        for instance in self.instances:
+            if instance.version == version:
+                return True
+
+        return False
+
+    def __iter__(self):
+        return iter(self.instances)
+
+
+CACHE = os.path.join(os.path.expanduser("~"), ".renutil")
+REGISTRY = Registry(os.path.join(CACHE, "index.bin"))
+
+
 @contextmanager
 def cd(dir):
     prevdir = os.getcwd()
@@ -147,54 +216,9 @@ def assure_state():
         logger.debug("Cache directory is not writeable:\n{}\nPlease make sure this script has permission to write to this directory.".format(CACHE))  # noqa: E501
         sys.exit(1)
     instances = scan_instances(CACHE)
-    if not os.path.isfile(INSTANCE_REGISTRY):
-        logger.debug("Instance registry does not exist, creating it:\n{}".format(INSTANCE_REGISTRY))
-        with open(INSTANCE_REGISTRY, "w") as f:
-            f.write(jsonpickle.encode(instances))
-    else:
-        for instance in instances:
-            add_to_registry(instance)
-
-
-def get_registry(args=None, unknown=None):
-    file = open(INSTANCE_REGISTRY, "r")
-    try:
-        registry = jsonpickle.decode(file.read())
-    except JSONDecodeError:
-        os.remove(INSTANCE_REGISTRY)
-        assure_state()
-        return None
-    return registry
-
-
-def remove_from_registry(instance):
-    registry = get_registry()
-    for i, inst in enumerate(registry):
-        if inst.version == instance.version:
-            del registry[i]
-    with open(INSTANCE_REGISTRY, "w") as f:
-        f.write(jsonpickle.encode(registry))
-
-
-def add_to_registry(instance):
-    registry = get_registry()
-    if instance not in registry:
-        registry.append(instance)
-        with open(INSTANCE_REGISTRY, "w") as f:
-            f.write(jsonpickle.encode(registry))
-
-
-def get_instance(version):
-    if isinstance(version, str):
-        try:
-            version = Version(version)
-        except ValueError:
-            return None
-    registry = get_registry()
-    for instance in registry:
-        if instance.version == version:
-            return instance
-    return None
+    REGISTRY.clear()
+    for instance in instances:
+        REGISTRY.add_instance(instance)
 
 
 def valid_version(version):
@@ -203,8 +227,8 @@ def valid_version(version):
             version = Version(version)
         except ValueError:
             return False
-    registry = get_registry()
-    for instance in registry:
+    # registry = get_registry()
+    for instance in REGISTRY:
         if instance.version == version:
             return True
     releases = get_available_versions()
@@ -240,15 +264,13 @@ def get_available_versions(args=None, unknown=None):
 
 
 def get_installed_versions(args=None, unknown=None):
-    return sorted(get_registry(), reverse=True)
+    return sorted(REGISTRY, reverse=True)
 
 
 @click.group(cls=AliasedGroup)
-@click.option("-d/-nd", "--debug/--no-debug", default=False,
-              help="Print debug information or only regular output")
+@click.option("-d", "--debug", is_flag=True)
 def cli(debug):
-    """
-    Commands can be abbreviated by the shortest unique string.
+    """Commands can be abbreviated by the shortest unique string.
 
     \b
     For example:
@@ -265,8 +287,7 @@ def cli(debug):
 @click.option("-n", "--num-versions", "count", default=5, type=int,
               help="Amount of versions to show, sorted in descending order")
 def list(show_all, count):
-    """
-    List all available versions of Ren'Py.
+    """List all available versions of Ren'Py.
     """
     assure_state()
     if show_all:
@@ -285,18 +306,6 @@ def list(show_all, count):
                 click.echo(release.version)
 
 
-def installed(version):
-    if isinstance(version, str):
-        try:
-            version = Version(version)
-        except ValueError:
-            return False
-    for instance in get_registry():
-        if instance.version == version:
-            return True
-    return False
-
-
 def download(url, dest):
     response = requests.head(url)
     if response.status_code == 404:
@@ -357,19 +366,39 @@ def get_members_tar(tar):
             yield tarinfo
 
 
+def patch_file(file, target_line, patch, reverse=False):
+    with open(file, "r") as f:
+        data = f.readlines()
+
+    found = False
+    with open(file, "w") as f:
+        for line in data:
+            if found:
+                if reverse:
+                    f.write(line)
+                    f.write(patch)
+                else:
+                    f.write(patch)
+                    f.write(line)
+                found = False
+            else:
+                f.write(line)
+            if target_line in line:
+                found = True
+
+
 @cli.command()
 @click.argument("version", required=True, type=str)
 @click.option("-f", "--force", is_flag=True)
 def install(version, force):
-    """
-    Install the specified version of Ren'Py (including RAPT).
+    """Install the specified version of Ren'Py (including RAPT).
     """
     assure_state()
-    if installed(version):
+    if REGISTRY.installed(version):
         if force:
             logger.info("Uninstalling {} before reinstalling...".format(version))
-            instance = get_instance(version)
-            remove_from_registry(instance)
+            instance = REGISTRY.get_instance(version)
+            REGISTRY.remove_instance(instance)
             shutil.rmtree(os.path.join(CACHE, instance.path))
             logger.info("Done uninstalling")
         else:
@@ -383,72 +412,73 @@ def install(version, force):
     sdk_filename = "renpy-{}-sdk.zip".format(version)
     rapt_filename = "renpy-{}-rapt.zip".format(version)
 
-    r = requests.get("https://www.renpy.org/dl/{}".format(version))
-
-    PYGAME_URL = None
-    soup = BeautifulSoup(r.content, "html.parser")
-    for link in soup.find_all("a"):
-        href = link.get("href")
-        if href.startswith("pygame_sdl2"):
-            pygame_filename = href
-            break
-
-    if not pygame_filename:
-        logger.error("Could not find pygame_sdl2 package.")
-        sys.exit(1)
-
     SDK_URL = "https://www.renpy.org/dl/{}/{}".format(version, sdk_filename)
     RAPT_URL = "https://www.renpy.org/dl/{}/{}".format(version, rapt_filename)
-    PYGAME_URL = "https://www.renpy.org/dl/{}/{}".format(version, pygame_filename)
 
     download(SDK_URL, os.path.join(CACHE, sdk_filename))
     download(RAPT_URL, os.path.join(CACHE, rapt_filename))
-    download(PYGAME_URL, os.path.join(CACHE, pygame_filename))
 
     logger.info("Extracting files...")
     sdk_zip = ZipFile(os.path.join(CACHE, sdk_filename), "r")
     rapt_zip = ZipFile(os.path.join(CACHE, rapt_filename), "r")
-    pygame_tar = tarfile.open(os.path.join(CACHE, pygame_filename), "r")
     sdk_zip.extractall(path=os.path.join(CACHE, version), members=get_members_zip(sdk_zip))
     rapt_zip.extractall(path=os.path.join(CACHE, version, "rapt"), members=get_members_zip(rapt_zip))
-    pygame_tar.extractall(path=os.path.join(CACHE, version, "pygame_sdl2"), members=get_members_tar(pygame_tar))
-
-    logger.info("Installing pygame_sdl2...")
-    pygame_path = os.path.join(CACHE, version, "pygame_sdl2")
-    if platform.mac_ver()[0]:
-        os.environ["MACOSX_DEPLOYMENT_TARGET"] = platform.mac_ver()[0]
-    with cd(pygame_path):
-        install = Popen(["python2", "setup.py", "install"], stdout=PIPE, stderr=PIPE)
-        for line in install.stdout:
-            logger.debug(str(line.strip(), "utf-8"))
-        install.communicate()
-        if install.returncode != 0:
-            logger.error("Could not install pygame_sdl2. You may need to install:")
-            logger.error("Linux: sudo apt install libsdl2-dev libpng-dev")
-            logger.error("macOS: brew install sdl2 sdl2_image sdl2_mixer sdl2_ttf libpng")
 
     logger.info("Installing RAPT...")
-    os.environ["PGS4A_NO_TERMS"] = "no"
     rapt_path = os.path.join(CACHE, version, "rapt")
+
+    arch = get_platform()
+    if arch == "windows-i686":
+        python_path = os.path.join(CACHE, version, "lib", arch, "python.exe")
+    else:
+        python_path = os.path.join(CACHE, version, "lib", arch, "python")
+        os.chmod(python_path, S_IRUSR | S_IXUSR)
+    site_package_path = os.path.join(CACHE, version, "lib", arch, "lib", "python2.7")
+
     with cd(rapt_path):
-        echo = Popen(["echo", """Y
-Y
-Y
-renutil"""], stdout=PIPE)
-        install = Popen(["python2", "android.py", "installsdk"], stdin=echo.stdout, stdout=PIPE)
+        patch = "sys.path.insert(0, '{}')\n\nimport ssl\nssl._create_default_https_context = ssl._create_unverified_context\n".format(site_package_path)  # noqa: E501
+        patch_file("android.py", "import sys", patch, reverse=True)
+
+        patch_file(os.path.join("buildlib", "rapt", "interface.py"),
+                   "def yesno_choice(self, prompt, default=None):",
+                   "        return True\n")
+
+        patch_file(os.path.join("buildlib", "rapt", "interface.py"),
+                   "def input(self, prompt, empty=None):",
+                   "        return \"renutil\"\n")
+
+    os.environ["RAPT_NO_TERMS"] = "no"
+    with cd(rapt_path):
+        logger.debug("Running {}".format(" ".join((python_path, "-O", "android.py", "installsdk"))))
+        install = Popen([python_path, "-O", "android.py", "installsdk"], stdout=PIPE, stderr=STDOUT)
         for line in install.stdout:
-            logger.debug(str(line.strip(), "utf-8"))
-    del os.environ["PGS4A_NO_TERMS"]
+            line = line.strip()
+            if line:
+                logger.debug(str(line, "latin-1"))
+    del os.environ["RAPT_NO_TERMS"]
 
     logger.info("Registering instance...")
     instance = RenpyInstance(version, version)
-    add_to_registry(instance)
+    REGISTRY.add_instance(instance)
 
     head, _ = os.path.split(get_libraries(instance)[0])
-    paths = [os.path.join(head, "python"), os.path.join(head, "pythonw"),
-             os.path.join(head, "renpy"), os.path.join(head, "zsync"), os.path.join(head, "zsyncmake")]
-    for path in paths:
-        os.chmod(path, S_IRUSR | S_IXUSR)
+    if arch != "windows-i686":
+        paths = [os.path.join(head, "python"), os.path.join(head, "pythonw"),
+                 os.path.join(head, "renpy"), os.path.join(head, "zsync"),
+                 os.path.join(head, "zsyncmake"),
+                 os.path.join(CACHE, instance.rapt_path, "project", "gradlew")]
+        for path in paths:
+            os.chmod(path, S_IRUSR | S_IXUSR)
+
+    with open(os.path.join(CACHE, instance.rapt_path, "project", "gradle.properties"), "r") as f:
+        original_content = f.readlines()
+
+    with open(os.path.join(CACHE, instance.rapt_path, "project", "gradle.properties"), "w") as f:
+        for line in original_content:
+            if line.startswith("org.gradle.jvmargs"):
+                f.write("org.gradle.jvmargs=-Xmx8g\n")
+            else:
+                f.write(line)
 
     logger.info("Done installing {}".format(version))
 
@@ -456,39 +486,53 @@ renutil"""], stdout=PIPE)
 @cli.command()
 @click.argument("version", required=True, type=str)
 def uninstall(version):
-    """
-    Uninstall the specified Ren'Py version.
+    """Uninstall the specified Ren'Py version.
     """
     assure_state()
-    if not installed(version):
+    if not REGISTRY.installed(version):
         logger.error("{} is not installed!".format(version))
         sys.exit(1)
-    instance = get_instance(version)
-    remove_from_registry(instance)
+    instance = REGISTRY.get_instance(version)
+    REGISTRY.remove_instance(instance)
     shutil.rmtree(os.path.join(CACHE, instance.path))
 
 
+def get_platform():
+    logger.debug("System: {}".format(platform.system()))
+    logger.debug("Machine: {}".format(platform.machine()))
+
+    if "Darwin" in platform.system():
+        return "darwin-x86_64"
+    elif "Windows" in platform.system():
+        return "windows-i686"
+    elif "x86_64" in platform.machine() or "amd64" in platform.machine():
+        return "linux-x86_64"
+    elif re.match(r"i.*86", platform.machine()):
+        return "linux-i686"
+    elif "Linux" in platform.system():
+        return "linux-{}".format(platform.machine())
+
+
 def get_libraries(instance):
-    info = os.uname()
-    platform = "{}-{}".format(info.sysname, info.machine)
     root = instance.path
     root1 = root
     root2 = root
     lib = None
-    if "Darwin" in info.sysname:
-        platform = "darwin-x86_64"
+    arch = get_platform()
+
+    if arch == "darwin-x86_64":
         root1 = root + "/../Resources/autorun"
         root2 = root + "/../../.."
-    elif "x86_64" in info.machine or "amd64" in info.machine:
-        platform = "linux-x86_64"
+    elif arch == "windows-i686":
         root1 = root
         root2 = root
-    elif re.match(r"i.*86", info.machine):
-        platform = "linux-i686"
+    elif arch == "linux-x86_64":
         root1 = root
         root2 = root
-    elif "Linux" in info.sysname:
-        platform = "linux-{}".format(info.machine)
+    elif arch == "linux-i686":
+        root1 = root
+        root2 = root
+    elif arch.startswith("linux-"):
         root1 = root
         root2 = root
     else:
@@ -496,13 +540,16 @@ def get_libraries(instance):
         sys.exit(1)
 
     for folder in [root, root1, root2]:
-        lib = os.path.join(CACHE, folder, "lib", platform)
+        lib = os.path.join(CACHE, folder, "lib", arch)
         if os.path.isdir(lib):
             break
-    lib = os.path.join(lib, "renpy")
+    if arch == "windows-i686":
+        lib = os.path.join(lib, "renpy.exe")
+    else:
+        lib = os.path.join(lib, "renpy")
 
     if not lib:
-        logger.error("Ren'Py platform files not found in '{}'".format(os.path.join(root, "lib", platform)))
+        logger.error("Ren'Py platform files not found in '{}'".format(os.path.join(root, "lib", arch)))
 
     if "LD_LIBRARY_PATH" in os.environ and len(os.environ["LD_LIBRARY_PATH"]) != 0:
         os.environ["LD_LIBRARY_PATH"] = "{}:{}".format(lib, os.environ["LD_LIBRARY_PATH"])
@@ -520,8 +567,7 @@ def get_libraries(instance):
 @click.option("-d", "--direct", is_flag=True)
 @click.argument("args", nargs=-1, type=click.UNPROCESSED)
 def launch(version, direct, args):
-    """
-    Launch the specified version of Ren'Py.
+    """Launch the specified version of Ren'Py.
 
     If invoked with default arguments, starts the 'launcher' project,
     which results in starting up the regular GUI launcher interface.
@@ -540,12 +586,16 @@ def launch(version, direct, args):
     \b
     Build Android distributions for a project:
         renutil launch <version> android_build <path_to_project_directory> assembleRelease|installDebug
+
+    \b
+    Lint a project:
+        renutil launch -d <version> <path_to_project_directory> lint
     """
     assure_state()
-    if not installed(version):
+    if not REGISTRY.installed(version):
         logger.error("{} is not installed!".format(version))
         sys.exit(1)
-    instance = get_instance(version)
+    instance = REGISTRY.get_instance(version)
     os.environ["SDL_AUDIODRIVER"] = "dummy"
     cmd = get_libraries(instance)
     if not direct:
@@ -562,17 +612,17 @@ def launch(version, direct, args):
 @cli.command()
 @click.argument("version", required=True, type=str)
 def cleanup(version):
-    """
-    Clean temporary files of the specified Ren'Py version.
+    """Clean temporary files of the specified Ren'Py version.
     """
     assure_state()
-    if not installed(version):
+    if not REGISTRY.installed(version):
         logger.error("{} is not installed!".format(version))
         sys.exit(1)
-    instance = get_instance(version)
+    instance = REGISTRY.get_instance(version)
     paths = [os.path.join(instance.path, "tmp"),
              os.path.join(instance.rapt_path, "assets"),
-             os.path.join(instance.rapt_path, "bin")]
+             os.path.join(instance.rapt_path, "bin"),
+             os.path.join(instance.rapt_path, "project", "app")]
     for path in paths:
         if os.path.isdir(os.path.join(CACHE, path)):
             shutil.rmtree(os.path.join(CACHE, path))
